# -*- coding: utf-8 -*-
################################################################################
# phamt/py_core.py
# The core Python implementation of the PHAMT and THAMT types.
# By Noah C. Benson

import sys, math
from collections.abc import Mapping

# ==============================================================================
# Constants
# The various constants related to PHAMTs and THAMTs are defined here. These
# largely mirror those of the phamt.h header file, but because we do not deal
# with bit-flags in the Python version, the BITS data are absent.

# Some of the hash data comes from the sys.hash_info.
HASH_BITCOUNT = sys.hash_info[0]
if   HASH_BITCOUNT == 16:  PHAMT_ROOT_SHIFT = 1
elif HASH_BITCOUNT == 32:  PHAMT_ROOT_SHIFT = 2
elif HASH_BITCOUNT == 64:  PHAMT_ROOT_SHIFT = 4
elif HASH_BITCOUNT == 128: PHAMT_ROOT_SHIFT = 3
else: raise RuntimeError("Unsupported hash size")

# We use a constant shift of 5 throughout except at the root node (which can't
# generally be shifted at 5 due to how the bits line-up--it instead gets the
# number of leftover bits in the hash integer, which was defined above as
# PHAMT_ROOT_SHIFT).
PHAMT_NODE_SHIFT = 5
PHAMT_TWIG_SHIFT = 5

# Here we define some consequences of the above definitions, which we use later.
PHAMT_ROOT_FIRSTBIT = (HASH_BITCOUNT - PHAMT_ROOT_SHIFT)
PHAMT_ROOT_MAXCELLS = (1 << PHAMT_ROOT_SHIFT)
PHAMT_NODE_MAXCELLS = (1 << PHAMT_NODE_SHIFT)
PHAMT_TWIG_MAXCELLS = (1 << PHAMT_TWIG_SHIFT)
PHAMT_ANY_MAXCELLS  = (1 << PHAMT_TWIG_SHIFT) # Assuming twig is largest.
PHAMT_NODE_BITS     = (HASH_BITCOUNT-PHAMT_ROOT_SHIFT-PHAMT_TWIG_SHIFT)
PHAMT_NODE_LEVELS   = (PHAMT_NODE_BITS // PHAMT_NODE_SHIFT)
PHAMT_LEVELS        = (PHAMT_NODE_LEVELS + 2) # (nodes + root + twig)
PHAMT_ROOT_DEPTH    = 0
PHAMT_TWIG_DEPTH    = (PHAMT_ROOT_DEPTH + PHAMT_NODE_LEVELS + 1)
PHAMT_LEAF_DEPTH    = (PHAMT_TWIG_DEPTH + 1)
PHAMT_ROOT_MASK     = ((1 << PHAMT_ROOT_SHIFT) - 1)
PHAMT_NODE_MASK     = ((1 << PHAMT_NODE_SHIFT) - 1)
PHAMT_TWIG_MASK     = ((1 << PHAMT_TWIG_SHIFT) - 1)
PHAMT_MAX_SHIFT     = max(PHAMT_TWIG_SHIFT, PHAMT_NODE_SHIFT, PHAMT_ROOT_SHIFT)
PHAMT_NCELLS        = (1 << PHAMT_MAX_SHIFT)

# These are used to ensure that a Python integer is represented as an unsigned
# integer--e.g., if you pass the key -1, this should translate to the maximum
# unsigned <HASH_BITCOUNT>-bit integer.
PHAMT_KEY_MIN = -(1 << (sys.hash_info[0] - 1))
PHAMT_KEY_MAX =  (1 << (sys.hash_info[0] - 1)) - 1
PHAMT_KEY_MOD = (1 << sys.hash_info[0])


# ==============================================================================
# Private Functions

def _highbitdiff(a, b):
    return int(math.log(a ^ b, 2))
def _key_to_hash(k):
    if not isinstance(k, int) or k > PHAMT_KEY_MAX or k < PHAMT_KEY_MIN:
        raise KeyError(k)
    return k if k >= 0 else (PHAMT_KEY_MOD + k)
def _depth_to_bit0shift(depth):
    if depth == 0:                return (PHAMT_ROOT_FIRSTBIT, PHAMT_ROOT_SHIFT)
    if depth == PHAMT_TWIG_DEPTH: return (0, PHAMT_TWIG_SHIFT)
    return (PHAMT_ROOT_FIRSTBIT - PHAMT_NODE_SHIFT*depth,
            PHAMT_NODE_SHIFT)
def _key_to_index(self, k):
    h = _key_to_hash(k)
    (bit0,shift) = self._b0sh
    a0 = bit0 + shift
    if (self._address >> a0) != (h >> a0): raise KeyError(k)
    return (h >> bit0) & ((1 << shift) - 1)
def _index_to_key(addr, ii):
    addr = addr | ii
    if addr > PHAMT_KEY_MAX: return addr - PHAMT_KEY_MOD
    else:                    return addr
def _phamt_from_kv(k, v, transient=False):
    h = _key_to_hash(k)
    addr = h & ~PHAMT_TWIG_MASK
    ii = h & PHAMT_TWIG_MASK
    cells = [None]*PHAMT_NCELLS
    cells[ii] = (v,)
    if not transient: cells = tuple(cells)
    return PHAMT(addr, PHAMT_TWIG_DEPTH, 1, cells)
def _phamt_join_disjoint(a, b, transient=False):
    a_addr = a._address
    b_addr = b._address
    hbd = _highbitdiff(a_addr, b_addr)
    if hbd >= HASH_BITCOUNT - PHAMT_ROOT_SHIFT:
        depth = 0
        bit0 = PHAMT_ROOT_FIRSTBIT
        shift = PHAMT_ROOT_SHIFT
    else:
        tmp = (hbd - PHAMT_TWIG_SHIFT) // PHAMT_NODE_SHIFT
        depth = PHAMT_LEVELS - 2 - tmp
        bit0 = tmp * PHAMT_NODE_SHIFT + PHAMT_TWIG_SHIFT
        shift = PHAMT_NODE_SHIFT
    addr = a_addr & ~((1 << (bit0 + shift)) - 1)
    cells = [None]*PHAMT_NCELLS
    tmp = ((1 << shift) - 1)
    a_ii = (a_addr >> bit0) & tmp
    b_ii = (b_addr >> bit0) & tmp
    cells[a_ii] = a
    cells[b_ii] = b
    numel = a._numel + b._numel
    if not transient: cells = tuple(cells)
    return PHAMT(addr, depth, numel, cells)
def _thamt_empty():
    return PHAMT(0, PHAMT_ROOT_DEPTH, 0, [None]*PHAMT_NCELLS)
def _thamt_set(self, k, v):
    if k > PHAMT_KEY_MAX or k < PHAMT_KEY_MIN: raise KeyError(k)
    try: ii = _key_to_index(self, k)
    except KeyError:
        # This key is outside the scope of this node, so we return a new
        # node that has this node and a new twig as children.
        twig = _phamt_from_kv(k, v, transient=True)
        return _phamt_join_disjoint(self, twig, transient=True)
    addr = self._address
    numel = self._numel
    depth = self._depth
    cells = self._cells
    if isinstance(cells, list):
        # This is a THAMT; we can edit in-place.
        cell = cells[ii]
        if depth == PHAMT_TWIG_DEPTH:
            cells[ii] = (v,)
            if cell is None: object.__setattr__(self, '_numel', numel + 1)
        elif cell is None:
            newcell = _phamt_from_kv(k, v, transient=True)
            if numel == 0: return newcell
            cells[ii] = newcell
            object.__setattr__(self, '_numel', numel + 1)
        else:
            oldn = cell._numel
            newcell = _thamt_set(cell, k, v)
            cells[ii] = newcell
            object.__setattr__(self, '_numel', numel - oldn + len(newcell))
        return self
    else:
        # This is a PHAMT; we return a new edited THAMT.
        cell = cells[ii]
        newcells = list(cells)
        if depth == PHAMT_TWIG_DEPTH:
            newcell = (v,)
            if cell is None: newnumel = numel + 1
            else:            newnumel = numel
        elif cell is None:
            newcell = _phamt_from_kv(k, v, transient=True)
            if numel == 0: return newcell
            newnumel = numel + 1
        else:
            oldn = cell._numel
            newcell = _thamt_set(cell, k, v)
            newnumel = numel - oldn + len(newcell)
        newcells[ii] = newcell
        return PHAMT(addr, depth, newnumel, newcells)
def _thamt_del(self, k):
    if k > PHAMT_KEY_MAX or k < PHAMT_KEY_MIN: raise KeyError(k)
    ii = _key_to_index(self, k)
    cells = self._cells
    cell = cells[ii]
    if cell is None: raise KeyError(k)
    addr = self._address
    numel = self._numel
    depth = self._depth
    istransient = isinstance(cells, list)
    newcells = cells if istransient else list(cells)
    if depth == PHAMT_TWIG_DEPTH:
        if numel == 1: return _thamt_empty()
        newcells[ii] = None
    else:
        newcell = _thamt_del(cell, k)
        if newcell._numel == 0:
            newcells[ii] = None
            # Figure out if we now have only 1 child:
            c0 = None
            c1 = None
            for (jj,c) in enumerate(newcells):
                if c is not None:
                    if c0 is None:
                        c0 = c
                    else:
                        c1 = c
                        break
            if c1 is None: return c0
        else:
            newcells[ii] = newcell
    if istransient:
        object.__setattr__(self, '_numel', numel - 1)
        return self
    else:
        return PHAMT(addr, depth, numel - 1, newcells)


# PHAMT Class ==================================================================

class PHAMT(Mapping):
    """A Persistent Hash Array Mapped Trie (PHAMT) type.                       
    
    The `PHAMT` class represents a minimal immutable persistent mapping type
    that can be used to implement persistent collections in Python
    efficiently. A `PHAMT` object is essentially a persistent dictionary that
    requires that all keys be Python integers (hash values); values may be any
    Python objects. `PHAMT` objects are highly efficient at storing either
    sparse hash values or lists of consecutive hash values, such as when the
    keys `0`, `1`, `2`, etc. are used.
    
    To add or remove key/valye pairs from a `PHAMT`, the methods
    `phamt_obj.assoc(k, v)` and `phamt_obj.dissoc(k)`, both of which return
    copies of `phamt_obj` with the requested change.
    
    `PHAMT` objects can be created in the following ways:
     * by using `phamt_obj.assoc(k,v)` or `phamt_obj.dissoc(k)` on existing
       `PHAMT` objects, such as the `PHAMT.empty` object, which represents
       an empty `PHAMT`;
     * by supplying the `PHAMT.from_list(iter_of_values)` with a list of
       values, which are assigned the keys `0`, `1`, `2`, etc.

    `PHAMT` objects should *not* be made by calling the `PHAMT` constructor.
    """
    empty = None
    __slots__ = ('_address', '_depth', '_b0sh', '_numel', '_cells')
    # The public interface.
    def __new__(cls, address, depth, numel, cells):
        self = super(PHAMT,cls).__new__(cls)
        object.__setattr__(self, '_address', address)
        object.__setattr__(self, '_depth', depth)
        object.__setattr__(self, '_b0sh', _depth_to_bit0shift(depth))
        object.__setattr__(self, '_numel', numel)
        object.__setattr__(self, '_cells', cells)
        return self
    def __setattr__(self, k, v):
        raise TypeError("type PHAMT is immutable")
    def __setitem__(self, k, v):
        # If cells is a list, this is a THAMT. #TODO
        raise TypeError("type PHAMT is immutable")
    def __getitem__(self, k):
        ii = _key_to_index(self, k)
        cells = self._cells
        c = cells[ii]
        if c is None: raise KeyError(k)
        if self._depth == PHAMT_TWIG_DEPTH: return c[0]
        else: return c[k]
    def __contains__(self, k):
        try: self.__getitem__(k)
        except KeyError: return False
        return True
    def __len__(self):
        return self._numel
    def __iter__(self):
        return PHAMTIter(self)
    def assoc(self, k, v):
        """Returns a new `PHAMT` object with an additional association.

        `phamt_obj.assoc(key, value)` returns a new `PHAMT` object that is equal
        to `phamt_obj` with the modification that in the new object, `key` is
        mapped to `value`. This is copied efficiently using shared state, so
        that the time to perform this update is `O(log n)` and the additional
        space required to keep both the original and the new object in memory is
        also `O(log n)`.
        """
        if k > PHAMT_KEY_MAX or k < PHAMT_KEY_MIN: raise KeyError(k)
        try: ii = _key_to_index(self, k)
        except KeyError:
            # This key is outside the scope of this node, so we return a new
            # node that has this node and a new twig as children.
            return _phamt_join_disjoint(self, _phamt_from_kv(k, v))
        addr = self._address
        numel = self._numel
        depth = self._depth
        cells = self._cells
        cell = cells[ii]
        newcells = list(cells)
        if depth == PHAMT_TWIG_DEPTH:
            newcell = (v,)
            if cell is None: newnumel = numel + 1
            else:            newnumel = numel
        elif cell is None:
            newcell = _phamt_from_kv(k, v)
            if numel == 0: return newcell
            newnumel = numel + 1
        else:
            newcell = cell.assoc(k, v)
            newnumel = numel - len(cell) + len(newcell)
        newcells[ii] = newcell
        return PHAMT(addr, depth, newnumel, tuple(newcells))
    def dissoc(self, k):
        """Returns a new `PHAMT` object with an additional association.

        `phamt_obj.assoc(key, value)` returns a new `PHAMT` object that is equal
        to `phamt_obj` with the modification that in the new object, `key` is
        mapped to `value`. This is copied efficiently using shared state, so
        that the time to perform this update is `O(log n)` and the additional
        space required to keep both the original and the new object in memory is
        also `O(log n)`.
        """
        if k > PHAMT_KEY_MAX or k < PHAMT_KEY_MIN: return self
        try: ii = _key_to_index(self, k)
        except KeyError: return self
        cells = self._cells
        cell = cells[ii]
        if cell is None: return self
        addr = self._address
        numel = self._numel
        depth = self._depth
        newcells = list(cells)
        if depth == PHAMT_TWIG_DEPTH:
            if numel == 1: return PHAMT.empty
            newcells[ii] = None
        else:
            newcell = cell.dissoc(k)
            if newcell is cell:
                return self
            elif newcell is PHAMT.empty:
                newcells[ii] = None
                # Figure out of we now have only 1 child:
                c0 = None
                c1 = None
                for (jj,c) in enumerate(newcells):
                    if c is not None:
                        if c0 is None:
                            c0 = c
                        else:
                            c1 = c
                            break
                if c1 is None: return c0
            else:
                newcells[ii] = newcell
        return PHAMT(addr, depth, numel - 1, tuple(newcells))
    def get(self, k, df):
        try:             return self.__getitem__(k)
        except KeyError: return df
    def transient(self):
        """Returns an equivalent transient HAMT (`THAMT`) object.

        `phamt.transient()` returns a transient `THAMT` object that is
        equivalent to `phamt`. This operation can be performed in constant time,
        and in-place updates to the resulting `THAMT` are performed with minimal
        allocations.
        """
        return THAMT(self)
    @staticmethod
    def from_iter(obj, k0=0):
        """Constructs a PHAMT object from a sequence or iterable of values.

        `PHAMT.from_iter(items)` returns a `PHAMT` object whose keys are the
        integers `0, 1 ... len(items)` and whose values are the elements of the
        iterable `items` in iteration order. This is performed with minimal
        allocations, so it should be more efficient than building the PHAMT from
        scratch.

        `PHAMT.from_iter(items, k0)` returns a `PHAMT` object whose keys are the
        integers `k0, k0+1 ... k0+len(items)`.
        """
        thamt = THAMT(PHAMT.empty)
        for obj in iter(obj):
            thamt[k0] = obj
            k0 += 1
        return thamt.persistent()
        
PHAMT.empty = PHAMT(0, PHAMT_ROOT_DEPTH, 0, (None,)*PHAMT_NCELLS)

# The PHAMT Iterator class.
class PHAMTIter(object):
    __slots__ = ('_phamt', '_stack')
    def __init__(self, phamt):
        self._phamt = phamt
        cells = phamt._cells
        self._stack = [(cells, None, phamt._depth, phamt._address)]
    def __iter__(self):
        return self
    def __next__(self):
        while len(self._stack) > 0:
            (top_cells, top_ii, top_depth, top_addr) = self._stack[-1]
            n = len(top_cells)
            if top_ii is None: top_ii = 0
            else:              top_ii += 1
            while top_ii < n and top_cells[top_ii] is None:
                top_ii += 1
            if top_ii >= n:
                self._stack.pop()
                continue
            cell = top_cells[top_ii]
            # We found a new element
            self._stack[-1] = (top_cells, top_ii, top_depth, top_addr)
            if top_depth == PHAMT_TWIG_DEPTH:
                return (_index_to_key(top_addr, top_ii), cell[0])
            else:
                self._stack.append(
                    (cell._cells, None, cell._depth, cell._address))
        raise StopIteration


# THAMT Class ==================================================================

class THAMT(object):
    """A Transient Hash Array Mapped Trie (THAMT) type.

    The `THAMT` class represents a minimal mutable persistent mapping type
    that can be used to efficiently edit persistent `PHAMT` objects in
    Python. (See also `PHAMT`). A `THAMT` object is essentially a tree
    structure that maps integer keys (hashes) to values, which may be any
    Python objects.

    `THAMT` objects can be greated from `PHAMT` objects (i.e.,
    `thamt = THAMT(phamt)`) in constant time. Unlike `PHAMT` objects,
    `THAMT` objects can be edited in-place like dictionaries. These edits
    are more efficient with respect to time than update to the `PHAMT` tyoe,
    however, they are slightly less space efficient than pure `PHAMT`s. Once
    a `THAMT` has been edited, it can be efficiently converted back into a
    `PHAMT` object using the `thamt.persistent()` method.
    """
    __slots__ = ('_phamt', '_version')
    def __init__(self, phamt=PHAMT.empty):
        if not isinstance(phamt, PHAMT):
            raise TypeError("can only make THAMTs from PHAMTs")
        object.__setattr__(self, '_phamt', phamt)
        object.__setattr__(self, '_version', 0)
    def __setattr__(self, k, v):
        raise TypeError("type THAMT does not allow attribute mutation")
    def __setitem__(self, k, v):
        u = _thamt_set(self._phamt, k, v)
        if u is not self._phamt:
            object.__setattr__(self, '_phamt', u)
        object.__setattr__(self, '_version', self._version + 1)
    def __delitem__(self, k):
        u = _thamt_del(self._phamt, k)
        if u is not self._phamt:
            object.__setattr__(self, '_phamt', u)
        object.__setattr__(self, '_version', self._version + 1)
    def __getitem__(self, k):
        return self._phamt.__getitem__(k)
    def __contains__(self, k):
        return self._phamt.__contains__(k)
    def __len__(self):
        return self._phamt._numel
    def __iter__(self):
        return THAMTIter(self)
    def get(self, k, nf):
        try: return self.__getitem__(k)
        except KeyError: return nf
    def persistent(self):
        if self._phamt._numel == 0: return PHAMT.empty
        # Basicaly, we crawl all the THAMTs turning them into PHAMTs then
        # return these newly-made PHAMTs.
        stack = [self._phamt]
        while len(stack) > 0:
            phamt = stack.pop()
            # Check this PHAMT.
            cells = phamt._cells
            if isinstance(cells, list):
                object.__setattr__(phamt, '_cells', tuple(cells))
                if phamt._depth == PHAMT_TWIG_DEPTH: continue
                for cell in cells:
                    if cell is not None:
                        stack.append(cell)
        return self._phamt

# The THAMT Iterator class.
class THAMTIter(object):
    __slots__ = ('_thamt', '_stack', '_version')
    def __init__(self, thamt):
        self._thamt = thamt
        self._version = thamt._version
        phamt = thamt._phamt
        cells = phamt._cells
        self._stack = [(cells, None, phamt._depth, phamt._address)]
    def __iter__(self):
        return self
    def __next__(self):
        if self._version != self._thamt._version:
            raise RuntimeError("THAMT changed during iteration")
        while len(self._stack) > 0:
            (top_cells, top_ii, top_depth, top_addr) = self._stack[-1]
            n = len(top_cells)
            if top_ii is None: top_ii = 0
            else:              top_ii += 1
            while top_ii < n and top_cells[top_ii] is None:
                top_ii += 1
            if top_ii >= n:
                self._stack.pop()
                continue
            cell = top_cells[top_ii]
            # We found a new element
            self._stack[-1] = (top_cells, top_ii, top_depth, top_addr)
            if top_depth == PHAMT_TWIG_DEPTH:
                return (_index_to_key(top_addr, top_ii), cell[0])
            else:
                self._stack.append(
                    (cell._cells, None, cell._depth, cell._address))
        raise StopIteration



# -*- coding: utf-8 -*-
####################################################################################################
# phamt/test/__init__.py
# Declaration of tests for the (C API) PHAMT type.
# By Noah C. Benson

import sys
from unittest import TestCase

class TestPHAMT(TestCase):
    """Tests of the `phamt.PHAMT` type."""

    MIN_INT = -(2 ** (sys.hash_info[0] - 1))
    MAX_INT = (2 ** (sys.hash_info[0] - 1) - 1)

    def make_random_pair(self, PHAMT, THAMT, n=100, minint=None, maxint=None):
        """Peforms a random set of operations on both a dict and a PHAMT and
        returns both.

        The dict is made with values that are weakrefs to the same objects that
        make up the PHAMT values. They should otherwise be equal.

        Also returns a list of (ordered) weakref references to the values that
        were inserted (some may be garbage collected already, that is fine).
        """
        from random import (randint, choice)
        from weakref import ref
        if minint is None: minint = TestPHAMT.MIN_INT
        if maxint is None: maxint = TestPHAMT.MAX_INT
        d = {}
        u = PHAMT.empty
        inserts = []
        for ii in range(n):
            # What to do? assoc or dissoc?
            if len(d) == 0 or randint(0,3):
                # assoc:
                v = set([ii])
                vr = ref(v)
                k = randint(minint, maxint)
                u = u.assoc(k, v)
                d[k] = vr
                inserts.append(vr)
            else:
                # Choose k frmo in d
                k = choice(list(d))
                u = u.dissoc(k)
                del d[k]
            # At every step, these must be equal.
            self.assertEqual(len(d), len(u))
            for (k,v) in d.items():
                self.assertTrue(k in u)
                self.assertTrue(v() is not None)
                self.assertTrue(u[k] is v())
        return (u, d, inserts)
    def make_random_pair_transient(self, PHAMT, THAMT,
                                   n=100, minint=None, maxint=None):
        """Peforms a random set of operations on both a dict and a THAMT and
        returns both.

        The dict is made with values that are weakrefs to the same objects that
        make up the THAMT values. They should otherwise be equal.

        Also returns a list of (ordered) weakref references to the values that
        were inserted (some may be garbage collected already, that is fine).

        At random intervals during its construction, the THAMT is persisted and
        checked against the dict.
        """
        from random import (randint, choice)
        from weakref import ref
        if minint is None: minint = TestPHAMT.MIN_INT
        if maxint is None: maxint = TestPHAMT.MAX_INT
        d = {}
        u = THAMT()
        inserts = []
        #lns = ["START"]
        for ii in range(n):
            # What to do? assoc or dissoc?
            if len(d) == 0 or randint(0,3):
                # assoc:
                v = set([ii])
                vr = ref(v)
                k = randint(minint, maxint)
                #lns.append(f"   assoc {k}")
                u[k] = v
                d[k] = vr
                inserts.append(vr)
            else:
                # Choose k from in d
                k = choice(list(d))
                #lns.append(f"   dissoc {k}")
                del u[k]
                del d[k]
            # Ocassionally, we persist and check the THAMT.
            try:
                if not randint(0, n // 10):
                    p = u.persistent()
                    self.assertEqual(len(d), len(p))
                    for (k,v) in d.items():
                        self.assertTrue(k in u)
                        self.assertTrue(v() is not None)
                        self.assertTrue(p[k] is v())
                # At every step, these must be equal.
                self.assertEqual(len(d), len(u))
                for (k,v) in d.items():
                    self.assertTrue(k in u)
                    self.assertTrue(v() is not None)
                    self.assertTrue(u[k] is v())
                for (k,v) in u:
                    self.assertTrue(k in d)
                    self.assertTrue(d[k]() is v)
            except Exception:
                #for ln in lns: print(ln)
                raise
        return (u, d, inserts)
    
    def pt_test_empty(self, PHAMT, THAMT):
        self.assertTrue(len(PHAMT.empty) == 0)
        self.assertFalse(any(x in PHAMT.empty for x in range(100)))
    def test_empty(self):
        """Tests that `PHAMT.empty` has the correct properties.

        `PHAMT.empty` is the empty `PHAMT`; it should have 0 length, and it
        should not contain anything.
        """
        from ..c_core import PHAMT, THAMT
        self.pt_test_empty(PHAMT, THAMT)
        from ..py_core import PHAMT, THAMT
        self.pt_test_empty(PHAMT, THAMT)
    def pt_test_iteration(self, PHAMT, THAMT):
        import gc
        (u, d, assocs) = self.make_random_pair(PHAMT, THAMT, 500)
        # First, iterate over d:
        n = 0
        for (k,v) in u:
            self.assertTrue(k in d)
            self.assertEqual(d[k](), v)
            n += 1
        self.assertEqual(n, len(d))
        # if we delete u, all the assocs should get gc'ed (i.e., iteration
        # shouldn't interfe with garbage collection).
        del u
        (k,v) = (None,None)
        gc.collect()
        for r in assocs:
            self.assertEqual(r(), None)
    def test_iteration(self):
        """Tests that `PHAMT` iteration works correctly.
        """
        from ..c_core import PHAMT, THAMT
        self.pt_test_iteration(PHAMT, THAMT)
        from ..py_core import PHAMT, THAMT
        self.pt_test_iteration(PHAMT, THAMT)
    def pt_test_edit(self, PHAMT, THAMT):
        """Tests that `PHAMT.assoc` and `PHAMT.dissoc` work correctly.

        The `assoc` method is used to add items to the PHAMT or to replace them.
        The `dissoc` method is used to remove items.
        """
        nought = PHAMT.empty
        p1 = nought.assoc(10, '10')
        p2 = p1.assoc(2000, '2000')
        p3 = p2.assoc(-50000, '-50000')
        # They should all have (and retain) correct lengths.
        self.assertTrue(len(nought) == 0)
        self.assertTrue(len(p1) == 1)
        self.assertTrue(len(p2) == 2)
        self.assertTrue(len(p3) == 3)
        # They should all have the keys they were given.
        self.assertTrue(10 not in nought and
                        10 in p1 and
                        10 in p2 and
                        10 in p3)
        self.assertTrue(2000 not in nought and
                        2000 not in p1 and
                        2000 in p2 and
                        2000 in p3)        
        self.assertTrue(-50000 not in nought and
                        -50000 not in p1 and
                        -50000 not in p2 and
                        -50000 in p3)
        # They should all have the correct values also.
        self.assertTrue(p1[10] == '10' and p2[10] == '10' and p3[10] == '10')
        self.assertTrue(p2[2000] == '2000' and p3[2000] == '2000')
        self.assertTrue(p3[-50000] == '-50000')
        # When looking up invalid values, they should raise errors.
        with self.assertRaises(KeyError):
            nought[10]
        with self.assertRaises(KeyError):
            p1[2000]
        with self.assertRaises(KeyError):
            p2[-50000]
        # Let's assoc up something more complex and see how it does.
        p = nought
        n = 100000
        for k in range(n):
            p = p.assoc(k, str(k))
        # Make sure the length is correct and it contains its items.
        self.assertEqual(len(p), n)
        for k in range(n):
            self.assertTrue(k in p)
            self.assertTrue(p[k] == str(k))
        # A more complicated (random) test that includes dissoc.
        (u, d, assocs) = self.make_random_pair(PHAMT, THAMT, 1000)
        self.assertTrue(len(u) == len(d))
        for (k,v) in d.items():
            self.assertTrue(k in u)
            self.assertTrue(u[k] is d[k]())
    def test_edit(self):
        """Tests that `PHAMT.assoc` and `PHAMT.dissoc` work correctly.

        The `assoc` method is used to add items to the PHAMT or to replace them.
        The `dissoc` method is used to remove items.
        """
        from ..c_core import PHAMT, THAMT
        self.pt_test_edit(PHAMT, THAMT)
        from ..py_core import PHAMT, THAMT
        self.pt_test_edit(PHAMT, THAMT)
    def pt_test_thamt(self, PHAMT, THAMT):
        import gc
        for k in range(5):
            (u, d, assocs) = self.make_random_pair_transient(PHAMT, THAMT, 500)
            # First, iterate over d::
            n = 0
            for (k,v) in u:
                self.assertTrue(k in d)
                self.assertEqual(d[k](), v)
                n += 1
            self.assertEqual(n, len(d))
            # Then check that these stay equal after persisting u.
            p = u#.persistent()
            n = 0
            for (k,v) in u:
                self.assertTrue(k in d)
                self.assertEqual(d[k](), v)
                n += 1
            self.assertEqual(n, len(d))
            n = 0
            for (k,v) in p:
                self.assertTrue(k in d)
                self.assertEqual(d[k](), v)
                n += 1
            self.assertEqual(n, len(d))
            # if we delete u, all the assocs should get gc'ed (i.e., iteration
            # shouldn't interfere with garbage collection).
            del u
            del p
            (k,v) = (None,None)
            gc.collect()
            for r in assocs:
                self.assertEqual(r(), None)
    def test_thamt(self):
        """Tests that THAMT objects work correctly.
        """
        from ..c_core import PHAMT, THAMT
        self.pt_test_thamt(PHAMT, THAMT)
        from ..py_core import PHAMT, THAMT
        self.pt_test_thamt(PHAMT, THAMT)
    def pt_test_gc(self, PHAMT, THAMT):
        import gc
        from weakref import ref
        s = set([])
        sr = ref(s)
        u1 = PHAMT.empty.assoc(0, s)
        del s
        self.assertTrue(sr() is not None)
        del u1
        gc.collect()
        self.assertTrue(sr() is None)
        # Test using random dict/PHAMT inserts and deletes.
        (u, d, assocs) = self.make_random_pair(PHAMT, THAMT, 500)
        # if we delete u, all the assocs should get gc'ed.
        del u
        gc.collect()
        for r in assocs:
            self.assertEqual(r(), None)
    def test_gc(self):
        """Tests that PHAMT objects are garbage collectable and allow their
        values to be garbage collected.
        """
        from ..c_core import PHAMT, THAMT
        self.pt_test_gc(PHAMT, THAMT)
        from ..py_core import PHAMT, THAMT
        self.pt_test_gc(PHAMT, THAMT)
    def test_from_iter(self):
        """Tests that PHAMT.from_iter works correctly.
        """
        from phamt.py_core import PHAMT as pyPHAMT
        from phamt.c_core import PHAMT as cPHAMT
        for PHAMT in [cPHAMT, pyPHAMT]:
            for (n,k0) in zip([10,50,100,5000],[0,-10,20,0]):
                u = PHAMT.from_iter(range(n), k0)
                d = {k0+ii:x for (ii,x) in enumerate(range(n))}
                self.assertEqual(len(u), len(d))
                for (k,v) in d.items():
                    self.assertTrue(k in u)
                    self.assertTrue(u[k] == v)
                for (k,v) in u:
                    self.assertTrue(d[k] == v)



# -*- coding: utf-8 -*-
################################################################################
# phamt/__init__.py
# Initialization file for the PHAMT library.
# By Noah C. Benson

"""Persistent and Transient Hash Array Mapped Trie data structures for Python.
"""

try:              from .c_core  import (PHAMT, THAMT)
except Exception: from .py_core import (PHAMT, THAMT)

__version__ = "0.1.2"




#! /usr/bin/env python
################################################################################

import os
from setuptools import (setup, Extension)
from setuptools.command.build_ext import build_ext
from sysconfig import get_config_var

# Depending on the C compiler, we have different options.
class BuildExt(build_ext):
    compile_flags = {"msvc": ["/EHsc", "/std:c11", "/GL"],
                     "unix": ["-std=c11", "-O3"]}
    def build_extensions(self):
        opts = self.compile_flags.get(self.compiler.compiler_type, [])
        for ext in self.extensions:
            ext.extra_compile_args = opts
        build_ext.build_extensions(self)

# Get the version from __init__.py
with open(os.path.join('phamt', '__init__.py'), 'rt') as fl:
    lns = fl.readlines()
version = next(ln for ln in lns if "__version__ = " in ln)
version = version.split('"')[1]

setup(
    name='phamt',
    version=version,
    description='C implementation of a Persistent Hash Array Mapped Trie',
    keywords='persistent immutable functional', 
    author='Noah C. Benson',
    author_email='nben@uw.edu',
    url='https://github.com/noahbenson/phamt/',
    download_url='https://github.com/noahbenson/phamt/',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: MacOS',
        'Programming Language :: C',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules'],
    packages=['phamt', 'phamt.test'],
    ext_modules=[
        Extension('phamt.c_core',
                  ['phamt/phamt.c'],
                  depends=['phamt/phamt.h'],
                  include_dirs=["phamt"],
                  language="c")],
    package_data={'': ['LICENSE.txt']},
    include_package_data=True,
    install_requires=[],
    cmdclass={'build_ext': BuildExt})
